/*
 *
 * X DLX Simulator
 *
 * The DLX is a 32-bit RISC machine based on the MIPS 2000 architecture.
 * The architecture includes 32 32-bit general purpose registers and a 
 * program counter.
 *
 * Floating point registers and an interrupt register are not implemented.
 *
 * The address space is 4 Gbytes, big-endian byte-addressable using 32-bit
 * addresses. All instructions must be properly aligned.
 *
 *
 * Cameron Humphries
 * Department of Computer Science
 * University of Adelaide
 * 1992
 *
 */
 
#include <stdio.h>
#include <signal.h>

#include <X11/X.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/Form.h>

#include "sim.h"

#define NO_INPUT 0
#define INPUT_OK 1

int input_allowed = NO_INPUT;

char *output_str=NULL;
int count_chars=0;
int num_reallocs=1;

extern XtAppContext app_context;
extern int sim_pid;

static void MyInsertChar();

static XtActionsRec act_table[] = {
  {"input-char",MyInsertChar},
};

char *translations =
"\
<Key>:		input-char() \n\
<FocusIn>:	focus-in() \n\
<FocusOut>:	focus-out() \n\
";

Widget text_view;

/*
 * 
 */
void Create_IO_View(Widget form)
{
  Widget w;
  static Arg label_args[] = {
    {XtNlabel, (XtArgVal)"INPUT/OUTPUT"},
    {XtNfromVert, (XtArgVal)NULL},
    {XtNfromHoriz, (XtArgVal)NULL},
    {XtNborderWidth, (XtArgVal)0},
  };
  static Arg text_args[] = {
    {XtNeditType, (XtArgVal)XawtextEdit},
    {XtNfromVert, (XtArgVal)NULL},
    {XtNfromHoriz, (XtArgVal)NULL},
    {XtNwidth, (XtArgVal)300},
    {XtNheight, (XtArgVal)350},
    {XtNstring, (XtArgVal)NULL},
    {XtNtype, (XtArgVal)XawAsciiString},
    {XtNscrollHorizontal, (XtArgVal)XawtextScrollAlways},
    {XtNscrollVertical, (XtArgVal)XawtextScrollAlways},
  };
  
  XtAppAddActions(app_context,act_table,XtNumber(act_table));


  /* Create title for I/O view */
  label_args[1].value = (XtArgVal)XtNameToWidget(form,"r31_label");
  label_args[2].value = (XtArgVal)XtNameToWidget(form,"Data_View");
  w = XtCreateManagedWidget("IO_Title",labelWidgetClass,form,
			    label_args,XtNumber(label_args));
			    
  /* Create READ-ONLY text widget */
  output_str = (char *)malloc(1001);
  
  text_args[1].value = (XtArgVal)w;
  text_args[2].value = (XtArgVal)XtNameToWidget(form,"Data_View");
  text_args[5].value = output_str;
  text_view = XtCreateManagedWidget("IO_View",asciiTextWidgetClass,form,
				    text_args,XtNumber(text_args));
  XtUninstallTranslations(text_view);
  XtOverrideTranslations(text_view,XtParseTranslationTable(translations));
  
  return;
}

static void MyInsertChar(Widget w,XEvent *event,String *p,Cardinal *n)
{
  char inp_char[10];
  int numchar=10,retval;
  KeySym keysym;
  static XComposeStatus status = {NULL, 0};
  
  retval=XLookupString(&event->xkey,inp_char,numchar,&keysym,&status);
  if (!retval)
    return;

  if (input_allowed)
    {
      if (retval)
	{
	  regs[1] = inp_char[0];
	  input_allowed = NO_INPUT;
	  kill(sim_pid,SIGUSR1);
	  Change_Text_Cursor();
	}
	
    }
  else
    {
      XBell(XtDisplay(w),50);
    }
  return;
}

void Prepare_Output(char ch)
{
  char *prev_nl,*ch_ptr;
  int i=0;
  int num_chars_for_lf;
  int insert_pos;
  int string_len;
    
  switch (ch)
    {
    case 7:
      XBell(XtDisplay(text_view),100);
      break;
    case 10:
      /* DO a line feed */
      insert_pos = count_chars;

      num_chars_for_lf=-1;
      for (i=insert_pos;i>0;i--)
	{
	  if (output_str[i] == '\n')
	    break;
	  num_chars_for_lf++;
	}

      count_chars = strlen(output_str);
      if ((num_reallocs*1000) < count_chars+num_chars_for_lf+1)
	output_str = (char *)realloc(output_str,(++num_reallocs*1000));
      output_str[count_chars++]='\n';
      for (i=0;i<num_chars_for_lf;i++)
	output_str[count_chars++]=' ';
      break;

    case 13:
      /* Do a carriage return */
      prev_nl = strrchr(output_str,'\n');

      if (!prev_nl)
	{
	  XawTextSetInsertionPoint(text_view,0); 
	  count_chars=0;
	}
      else
	{
	  ch_ptr = output_str;
	  while (ch_ptr != prev_nl)
	    {
	      i++;
	      ch_ptr++;
	    }
	  count_chars = i+1;
	  XawTextSetInsertionPoint(text_view,count_chars); 
	}
      break;

    default:
      /* Write out char */
      output_str[count_chars++] = ch;
      output_str[count_chars] = '\0';
      if ((num_reallocs*1000) == count_chars)
	output_str = (char *)realloc(output_str,(++num_reallocs*1000));
      break;
    }
    
  return;
}

void Reset_Output()
{
  static Arg arg_list[] = {
    {XtNstring,(XtArgVal)NULL},
  };

  free(output_str);
  output_str = (char *)malloc(1000);
  bzero(output_str,1000);
  arg_list[0].value = (XtArgVal)output_str;
  XtSetValues(text_view,arg_list,XtNumber(arg_list));
  count_chars = 0;
  num_reallocs = 1;
  XawTextSetInsertionPoint(text_view,count_chars);

  return;
}

void Change_Text_Cursor()
{
  static Arg text_args[] = {
    {XtNcursorName,(XtArgVal)"hand2"},
  };
  
  if (input_allowed)
    {
      text_args[0].value = (XtArgVal)"hand2";
      XtSetValues(text_view,text_args,XtNumber(text_args));
    }
  else
    {
      text_args[0].value = (XtArgVal)"xterm";
      XtSetValues(text_view,text_args,XtNumber(text_args));
    }
    
  return;
}

void Display_Output()
{
  static Arg arg_list[] = {
    {XtNstring,(XtArgVal)NULL},
  };

  arg_list[0].value = (XtArgVal)output_str;
  XtSetValues(text_view,arg_list,XtNumber(arg_list));
  XawTextSetInsertionPoint(text_view,count_chars);
}
